==> GotoBLAS-1.00.tar.gz: GotoBLAS/common_thread.h <==
/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#ifndef COMMON_THREAD
#define COMMON_THREAD

/* Thread Timing Debugging flag */
#undef TIMING_DEBUG

/* Global Parameter */
extern int blas_cpu_number;

#ifndef USE_BUSY_WAIT
#define VOLATILE
#else
#define VOLATILE volatile
#endif

#define BLAS_SINGLE	0x0000U
#define BLAS_DOUBLE	0x0001U
#define BLAS_REAL	0x0000U
#define BLAS_COMPLEX	0x0002U
#define BLAS_TRANSA	0x000CU	/* 2bit */
#define BLAS_TRANSA_N	0x0000U
#define BLAS_TRANSA_T	0x0004U
#define BLAS_TRANSA_R	0x0008U
#define BLAS_TRANSA_C	0x000CU
#define BLAS_TRANSA_SHIFT     2
#define BLAS_TRANSB	0x0030U	/* 2bit */
#define BLAS_TRANSB_N	0x0000U
#define BLAS_TRANSB_T	0x0010U
#define BLAS_TRANSB_R	0x0020U
#define BLAS_TRANSB_C	0x0030U
#define BLAS_TRANSB_SHIFT     4
#define BLAS_RSIDE      0x0040U
#define BLAS_RSIDE_SHIFT      6
#define BLAS_UPLO       0x0080U
#define BLAS_UPLO_SHIFT       7

#define BLAS_STATUS_NOTYET	0
#define BLAS_STATUS_QUEUED	1
#define BLAS_STATUS_RUNNING	2
#define BLAS_STATUS_FINISHED	4

struct blas_queue {
  pthread_mutex_t lock;
  pthread_cond_t finish;

  void *routine;  /* function pointer */
  int   mode;     /* double/single/real/complex/rside/lside ... */
  long m, n, k;
  void *a, *b, *c;
  long lda, ldb, ldc;
  double alpha[2];
  long from, to;
  int status;
  int last;

  struct blas_queue *next;

#ifdef SMP_DEBUG
  int    num;
#endif
#ifdef TIMING_DEBUG
  unsigned int clocks;
#endif
};

typedef struct blas_queue blas_queue_t;

#ifdef SMP_SERVER

extern int blas_server_avail;

#ifdef USE_BUSY_WAIT
typedef volatile struct {
  blas_param_t       *param;
  long dummypad[6];
} blas_each_param_t;
#endif

/* Thread server common infomation */
typedef struct{
  pthread_mutex_t	 lock;
  pthread_cond_t	 filled;   /* queue is filled */
  blas_queue_t		*queue;    /* Parameter Pointer */
  int			 shutdown; /* server shutdown flag */
} blas_pool_t;

static __inline void blas_pool_init(blas_pool_t *pool){

  pthread_mutex_init(&pool->lock, NULL);
  pthread_cond_init (&pool->filled, NULL);
  pool -> shutdown = 0;
  pool->queue = NULL;
}

static __inline void blas_queue_init(blas_queue_t *queue){

  pthread_mutex_init(&queue->lock, NULL);
  pthread_cond_init (&queue->finish, NULL);
  queue-> status = BLAS_STATUS_NOTYET;
  queue-> last   = 0;
  queue-> next  = NULL;

}

static __inline void blas_queue_destroy(blas_queue_t *queue){

  pthread_cond_destroy(&queue->finish);
  if (pthread_mutex_destroy(&queue->lock)) {
    fprintf(STDERR, "BLAS Thread Warning : MUTEX is released without beeing unlocked!\n");
  }

}

int blas_thread_init(void);
int blas_thread_shutdown_(void);
int exec_blas(blas_queue_t *, void *buffer);
int exec_blas_async(blas_queue_t *);
int exec_blas_async_wait(blas_queue_t *);
#ifdef USE_BUSY_WAIT
int blas_make_it_busy(int flag);
#endif
#else
int exec_blas_async(int num_cpu, blas_param_t *param, pthread_t *);
int exec_blas_async_wait(int num_cpu, pthread_t *blas_threads);
int exec_blas(int num_cpu, blas_param_t *param, void *buffer);
#endif

#ifndef ASSEMBLER

int blas_level1_thread(int mode, long m, long n, long k,
		       double alpha_r, double alpha_i,
		       void *a, long lda,
		       void *b, long ldb, 
		       void *c, long ldc, int (*function)());

int gemv_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *b, long ldb, 
		void *c, long ldc, int (*function)(), void *buffer);

int ger_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *b, long ldb, 
		void *c, long ldc, int (*function)(), void *buffer);

int gemm_thread(int mode, long m, long n, long k, 
		double alpha_r, double alpha_i,
		void *a, long lda, 
		void *b, long ldb, 
		void *c, long ldc, int (*function)(), void *buffer);

int trsm_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *c, long ldc, int (*function)(), void *buffer);

int symm_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *b, long ldb,
		void *c, long ldc, int (*function)(), void *buffer);

int syrk_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *b, long ldb,
		void *c, long ldc, int (*function)(),
		void *buffer);

int beta_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *c, long ldc, int (*fuction)());

int getrf_thread(int mode, long m, long n, long k,
		 void *offsetA, long lda,
		 void *offsetB, long jb,
		 void *ipiv, long offset, int (*function)(), void *buffer);

#ifdef SMP_DEBUG
static void show_param(blas_param_t *param){
  fprintf(STDERR, "------ Parameter ----------------\n");
  fprintf(STDERR, "mode = %2d\n", param->mode);
  fprintf(STDERR, "m = %3ld  n = %ld  k = %3ld\n", param->m, param->n, param->k);
  fprintf(STDERR, "A = 0x%016lx  lda = %3ld\n", 
	  (unsigned long)param->a, param->lda);
  fprintf(STDERR, "B = 0x%016lx  ldb = %3ld\n", 
	  (unsigned long)param->b, param->ldb);
  fprintf(STDERR, "C = 0x%016lx  ldc = %3ld\n", 
	  (unsigned long)param->c, param->ldc);
  fprintf(STDERR, "alpha_r = %e  alpha_i = %e\n",
	  param->alpha[0], param->alpha[1]);
  fprintf(STDERR, "--------------------------------\n");
}

static void show_mode(int mode){
  fprintf(STDERR, "---------- Mode ----------------\n");
  fprintf(STDERR, "DOUBLE  = %d\t", ((mode & BLAS_DOUBLE ) != 0));
  fprintf(STDERR, "COMPLEX = %d\n", ((mode & BLAS_COMPLEX) != 0));
  fprintf(STDERR, "TRANS A = %d\t", (mode >>BLAS_TRANSA_SHIFT) & 3);
  fprintf(STDERR, "TRANS B = %d\n", (mode >>BLAS_TRANSB_SHIFT) & 3);
  fprintf(STDERR, "RSIDE   = %d\t", (mode >>BLAS_RSIDE_SHIFT) & 1);
  fprintf(STDERR, "UPLO    = %d\n", (mode >>BLAS_UPLO_SHIFT) & 1);
  fprintf(STDERR, "--------------------------------\n");
}
#endif
#endif  /* ENDIF ASSEMBLER */

#endif

==> GotoBLAS-1.08.tar.gz: GotoBLAS/common_thread.h <==
/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#ifndef COMMON_THREAD
#define COMMON_THREAD

/* Basic Thread Debugging */
#undef SMP_DEBUG

/* Thread Timing Debugging */
#undef TIMING_DEBUG

/* Global Parameter */
extern int blas_cpu_number;

#ifndef USE_BUSYWAIT
#define VOLATILE
#else
#define VOLATILE volatile
#endif

#define BLAS_SINGLE	0x0000U
#define BLAS_DOUBLE	0x0001U
#define BLAS_REAL	0x0000U
#define BLAS_COMPLEX	0x0002U
#define BLAS_TRANSA	0x000CU	/* 2bit */
#define BLAS_TRANSA_N	0x0000U
#define BLAS_TRANSA_T	0x0004U
#define BLAS_TRANSA_R	0x0008U
#define BLAS_TRANSA_C	0x000CU
#define BLAS_TRANSA_SHIFT     2
#define BLAS_TRANSB	0x0030U	/* 2bit */
#define BLAS_TRANSB_N	0x0000U
#define BLAS_TRANSB_T	0x0010U
#define BLAS_TRANSB_R	0x0020U
#define BLAS_TRANSB_C	0x0030U
#define BLAS_TRANSB_SHIFT     4
#define BLAS_RSIDE      0x0040U
#define BLAS_RSIDE_SHIFT      6
#define BLAS_UPLO       0x0080U
#define BLAS_UPLO_SHIFT       7

#define BLAS_STATUS_NOTYET	0
#define BLAS_STATUS_QUEUED	1
#define BLAS_STATUS_RUNNING	2
#define BLAS_STATUS_FINISHED	4

struct blas_queue {
#if defined( __WIN32__) || defined(__CYGWIN32__)
  CRITICAL_SECTION lock;
  HANDLE finish;
#else
  pthread_mutex_t lock;
  pthread_cond_t finish;
#endif

  void *routine;  /* function pointer */
  int   mode;     /* double/single/real/complex/rside/lside ... */
  long m, n, k;
  void *a, *b, *c;
  long lda, ldb, ldc;
  double alpha[2];
  long from, to;
  int status;
  int last;

  struct blas_queue *next;

#ifdef SMP_DEBUG
  int    num;
#endif
#ifdef TIMING_DEBUG
  unsigned int clocks;
#endif
};

typedef struct blas_queue blas_queue_t;

#ifdef SMP_SERVER

extern int blas_server_avail;

/* Thread server common infomation */
typedef struct{
#if defined( __WIN32__) || defined(__CYGWIN32__)
  CRITICAL_SECTION lock;
  HANDLE filled;
  HANDLE killed;
#else
  pthread_mutex_t	 lock;
  pthread_cond_t	 filled;   /* queue is filled */
#endif

  blas_queue_t	*queue;    /* Parameter Pointer */

#ifdef USE_BUSYWAIT
  volatile int		 busywait;
  blas_queue_t *volatile busyqueue[MAX_CPU_NUMBER];
#endif

  int			 shutdown; /* server shutdown flag */
} blas_pool_t;

static __inline void blas_pool_init(blas_pool_t *pool){

#ifdef USE_BUSYWAIT
  int i;
#endif

#if defined( __WIN32__) || defined(__CYGWIN32__)
  InitializeCriticalSection(&pool->lock);
  pool->filled = CreateEvent(NULL, FALSE, FALSE, NULL);
  pool->killed = CreateEvent(NULL, TRUE,  FALSE, NULL);
#else
  pthread_mutex_init(&pool->lock, NULL);
  pthread_cond_init (&pool->filled, NULL);
#endif

  pool -> shutdown = 0;
  pool -> queue    = NULL;

#ifdef USE_BUSYWAIT
  pool -> busywait  = 0;
  for (i = 0; i < MAX_CPU_NUMBER; i++) pool -> busyqueue[i] = NULL;
#endif
}

static __inline void blas_queue_init(blas_queue_t *queue){

#if defined( __WIN32__) || defined(__CYGWIN32__)
  InitializeCriticalSection(&queue->lock);
  queue->finish = CreateEvent(NULL, FALSE, FALSE, NULL);
#else
  pthread_mutex_init(&queue->lock, NULL);
  pthread_cond_init (&queue->finish, NULL);
#endif
  queue-> status = BLAS_STATUS_NOTYET;
  queue-> last   = 0;
  queue-> next  = NULL;

}

static __inline void blas_queue_destroy(blas_queue_t *queue){

#if defined( __WIN32__) || defined(__CYGWIN32__)
  CloseHandle(&queue->finish);
  DeleteCriticalSection(&queue->lock);
#else
  pthread_cond_destroy(&queue->finish);
  if (pthread_mutex_destroy(&queue->lock)) {
    fprintf(STDERR, "BLAS Thread Warning : MUTEX is released without beeing unlocked!\n");
  }
#endif

}

int blas_thread_init(void);
int blas_thread_shutdown_(void);
int exec_blas(blas_queue_t *, void *buffer);
int exec_blas_async(blas_queue_t *);
int exec_blas_async_wait(blas_queue_t *);

#ifdef USE_BUSYWAIT
int blas_busywait(int flag);
#endif

#else
int exec_blas_async(int num_cpu, blas_param_t *param, pthread_t *);
int exec_blas_async_wait(int num_cpu, pthread_t *blas_threads);
int exec_blas(int num_cpu, blas_param_t *param, void *buffer);
#endif

#ifndef ASSEMBLER

int blas_level1_thread(int mode, long m, long n, long k,
		       double alpha_r, double alpha_i,
		       void *a, long lda,
		       void *b, long ldb, 
		       void *c, long ldc, int (*function)());

int gemv_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *b, long ldb, 
		void *c, long ldc, int (*function)(), void *buffer);

int ger_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *b, long ldb, 
		void *c, long ldc, int (*function)(), void *buffer);

int gemm_thread(int mode, long m, long n, long k, 
		double alpha_r, double alpha_i,
		void *a, long lda, 
		void *b, long ldb, 
		void *c, long ldc, int (*function)(), void *buffer);

int trsm_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *c, long ldc, int (*function)(), void *buffer);

int symm_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *b, long ldb,
		void *c, long ldc, int (*function)(), void *buffer);

int syrk_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *a, long lda,
		void *b, long ldb,
		void *c, long ldc, int (*function)(),
		void *buffer);

int beta_thread(int mode, long m, long n, 
		double alpha_r, double alpha_i,
		void *c, long ldc, int (*fuction)());

int getrf_thread(int mode, long m, long n, long k,
		 void *offsetA, long lda,
		 void *offsetB, long jb,
		 void *ipiv, long offset, int (*function)(), void *buffer);

#endif  /* ENDIF ASSEMBLER */

#endif
